use std::fs::File;
use std::path::PathBuf;
use std::io::{Write, BufRead, BufReader};

use color_eyre::Result;
use crossterm::event::{KeyCode, KeyEvent};
use ratatui::layout::{Constraint, Margin};
use ratatui::text::Text;
use ratatui::widgets::{Cell, Padding, Scrollbar, ScrollbarOrientation};
use ratatui::{
    layout::{Alignment, Rect}, 
    widgets::{Block, ScrollbarState, Row, Table, TableState}, 
    Frame
};
use subxt::ext::sp_core::crypto::{Ss58Codec, Ss58AddressFormat, AccountId32};

use super::{Component, PartialComponent, CurrentTab};
use crate::{
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

pub struct AddressBook {
    is_active: bool,
    address_book_file: PathBuf,
    address_book: Vec<(String, String, AccountId32, String)>,
    scroll_state: ScrollbarState,
    table_state: TableState,
    palette: StylePalette,
}

impl Default for AddressBook {
    fn default() -> Self {
        Self::new()
    }
}

impl AddressBook {
    pub fn new() -> Self {
        Self {
            is_active: false,
            address_book_file: Default::default(),
            address_book: Vec::new(),
            scroll_state: ScrollbarState::new(0),
            table_state: TableState::new(),
            palette: StylePalette::default(),
        }
    }

    fn save_to_file(&mut self) {
        let mut file = File::create(&self.address_book_file).unwrap();
        for wallet in self.address_book.iter() {
            writeln!(file, "{}:0x{}", wallet.0, &wallet.3).unwrap();
        }
    }

    fn read_or_create(&mut self, file_path: &PathBuf) -> Result<()> {
        assert!(self.address_book.len() == 0, "address_book already exists");
        match File::open(file_path) {
            Ok(file) => {
                let reader = BufReader::new(file);
                for line in reader.lines() {
                    let line = line?.replace("\n", "");
                    let line_split_at = line.find(":").unwrap_or(line.len());
                    let (name, seed) = line.split_at(line_split_at);
                    let seed_str = &seed[3..];

                    let seed: [u8; 32] = hex::decode(seed_str)
                        .expect("stored seed is valid hex string; qed")
                        .as_slice()
                        .try_into()
                        .expect("stored seed is valid length; qed");

                    let account_id = AccountId32::from(seed);
                    let address = AccountId32::from(seed.clone())
                        .to_ss58check_with_version(Ss58AddressFormat::custom(1996));
                    self.address_book.push((name.to_string(), address, account_id, seed_str.to_string()));
                }
            },
            Err(_) => {
                let chad_boyz = vec![
                    ("Pierre",   "328d3b7c3046ef7700937d99fb2e98ce2591682c2b5dcf3f562e4da157650237"),
                    ("Ghost_7",  "3666e4e19f87bb8680495f31864ce1f1c69d4178002cc01911aef2cc7313f203"),
                    ("Neptune1", "ac871e8bab00dd56ba3a1c0bd289357203dcaf10010b0b04ad7472870cd22a3c"),
                    ("Neptune2", "425ccd7bda4f5c76788ba23bc0381d7a2e496179c93301208c57501c80a4232a"),
                    ("Doctor K", "927a98dcf8f721103005f168476c24b91d7d10d580f457006a908e10e62c7729"),
                    ("Starman",  "ac9e227e30a63ce6eeb55cfbb1fb832aa7e1d3fad2bcb3f663de4a91d744fd50"),
                    ("Kitsune1", "46c78fcacffd80abc9cca4917ef8369a37e21a1691ca11e7a3b53f80be745313"),
                    ("Scientio", "fa5e5a295ec74c3dda81118d9240db1552b28f831838465ae0712e97e78a6728"),
                    ("Kitsune2", "4078ddb1ba1388f768fe6aa40ba9124a72692ecbcc83dc088fa86c735e4dc128"),
                    ("Proxmio",  "5e1456904c40192cd3a18183df7dffea90d97739830a902cabb702ecdae4f649"),
                ];

                let mut new_file = File::create(file_path)?;
                chad_boyz
                    .iter()
                    .for_each(|chad_info| {
                        writeln!(new_file, "{}:0x{}", chad_info.0, chad_info.1)
                            .expect("should write to address book; qed");
                        let chad_account_id: [u8; 32] = hex::decode(&chad_info.1[..])
                            .expect("stored seed is valid hex string; qed")
                            .as_slice()
                            .try_into()
                            .expect("stored seed is valid length; qed");
                        let chad_account_id = AccountId32::from(chad_account_id);
                        let address = AccountId32::from(chad_account_id.clone())
                            .to_ss58check_with_version(Ss58AddressFormat::custom(1996));
                        self.address_book.push((chad_info.0.to_string(), address, chad_account_id, chad_info.1.to_string()));
                    });
            }
        };
        self.scroll_state = self.scroll_state.content_length(self.address_book.len());
        Ok(())
    }

    fn swap_up(&mut self) {
        if let Some(src_index) = self.table_state.selected() {
            let dst_index = src_index.saturating_sub(1);
            if src_index > dst_index {
                self.address_book.swap(src_index, dst_index);
                self.previous_row();
                self.save_to_file();
            }
        }
    }

    fn swap_down(&mut self) {
        if let Some(src_index) = self.table_state.selected() {
            let dst_index = src_index + 1;
            if dst_index < self.address_book.len() {
                self.address_book.swap(src_index, dst_index);
                self.next_row();
                self.save_to_file();
            }
        }
    }

    fn delete_row(&mut self) {
        if let Some(index) = self.table_state.selected() {
            let _ = self.address_book.remove(index);
            self.previous_row();
            self.save_to_file();
        }
    }

    fn first_row(&mut self) {
        if self.address_book.len() > 0 {
            self.table_state.select(Some(0));
            self.scroll_state = self.scroll_state.position(0);
        }
    }

    fn next_row(&mut self) {
        let i = match self.table_state.selected() {
            Some(i) => {
                if i >= self.address_book.len() - 1 {
                    i
                } else {
                    i + 1
                }
            },
            None => 0,
        };
        self.table_state.select(Some(i));
        self.scroll_state = self.scroll_state.position(i);
    }

    fn last_row(&mut self) {
        if self.address_book.len() > 0 {
            let last = self.address_book.len() - 1;
            self.table_state.select(Some(last));
            self.scroll_state = self.scroll_state.position(last);
        }
    }

    fn previous_row(&mut self) {
        let i = match self.table_state.selected() {
            Some(i) => {
                if i == 0 {
                    0
                } else {
                    i - 1
                }
            },
            None => 0
        };
        self.table_state.select(Some(i));
        self.scroll_state = self.scroll_state.position(i);
    }
}

impl PartialComponent for AddressBook {
    fn set_active(&mut self, current_tab: CurrentTab) {
        match current_tab {
            CurrentTab::AddressBook => self.is_active = true,
            _ => {
                self.is_active = false;
                self.table_state.select(None);
                self.scroll_state = self.scroll_state.position(0);
            }
        }
    }
}

impl Component for AddressBook {
    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Explorer) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_hover_style(style.get("hover_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_hover_border_style(style.get("hover_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_hover_title_style(style.get("hover_title_style").copied());
            self.palette.with_highlight_style(style.get("highlight_style").copied());
            self.palette.with_scrollbar_style(style.get("scrollbar_style").copied());
        }

        let mut address_book_file = config.config.data_dir;
        address_book_file.push("address-book");
        self.read_or_create(&address_book_file)?;
        self.address_book_file = address_book_file;
        Ok(())
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        match key.code {
           KeyCode::Up | KeyCode::Char('k') if self.is_active => self.previous_row(), 
           KeyCode::Down | KeyCode::Char('j') if self.is_active => self.next_row(), 
           KeyCode::Char('g') if self.is_active => self.first_row(),
           KeyCode::Char('G') if self.is_active => self.last_row(),
           KeyCode::Char('K') if self.is_active => self.swap_up(),
           KeyCode::Char('J') if self.is_active => self.swap_down(),
           KeyCode::Char('D') if self.is_active => self.delete_row(),
           _ => {},
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, place] = super::bars_layout(area);
        let (border_style, border_type) = self.palette.create_border_style(self.is_active);

        let table = Table::new(
            self.address_book
                .iter()
                .map(|info| Row::new(vec![
                        Cell::from(Text::from(info.0.clone()).alignment(Alignment::Left)), 
                        Cell::from(Text::from(info.1.clone()).alignment(Alignment::Center)), 
                        Cell::from(Text::from("31 CSPR".to_string()).alignment(Alignment::Right)),
                ])),
            [
                Constraint::Min(5),
                Constraint::Max(51),
                Constraint::Min(10),
            ],
        )
        .style(self.palette.create_basic_style(false))
        .highlight_style(self.palette.create_basic_style(true))
        .column_spacing(1)
        .block(Block::bordered()
            .border_style(border_style)
            .border_type(border_type)
            .padding(Padding::right(2))
            .title_alignment(Alignment::Right)
            .title_style(self.palette.create_title_style(false))
            .title("Address Book"));

        let scrollbar = Scrollbar::default()
            .orientation(ScrollbarOrientation::VerticalRight)
            .begin_symbol(None)
            .end_symbol(None)
            .style(self.palette.create_scrollbar_style());

        frame.render_stateful_widget(table, place, &mut self.table_state);
        frame.render_stateful_widget(
            scrollbar, 
            place.inner(Margin { vertical: 1, horizontal: 1 }), 
            &mut self.scroll_state,
        );

        Ok(())
    }
}
